# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import re

from contrast.extern import webob

from contrast.api import dtm_pb2
from contrast.utils.timer import now_ms
from contrast.agent.middlewares.route_coverage.common import (
    get_normalized_uri as strip_uri,
    get_url_parameters as find_parameters,
)

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


class Request(webob.BaseRequest):
    def __init__(self, environ):
        """
        Django is not fully wsgi-compliant, so we need its request body to be passed in
        explicitly.
        """
        super().__init__(environ)

        self._document_type = None
        self._normalized_uri = None
        self._url_parameters = None

        self.timestamp_ms = now_ms()

    def get_dtm(self):
        """
        Builds the protobuf HttpRequest object from the current Request object
        :return: HttpRequest protobuf dtm
        """
        dtm = dtm_pb2.HttpRequest()

        dtm.uuid = str(id(self))
        dtm.timestamp_ms = self.timestamp_ms

        dtm.sender.ip = self.client_addr or ""
        dtm.sender.host = self.remote_host or ""

        # NOTE: we don't have a reliable way to extract dtm.receiver.ip from the WSGI environ
        dtm.receiver.host = self.domain
        dtm.receiver.port = int(self.host_port)

        dtm.protocol = self.scheme
        dtm.version = self._get_http_version()
        dtm.method = self.method
        dtm.raw = self.path_qs

        dtm.parsed_request_body = True
        dtm.request_body_binary = self.get_body()
        dtm.document_type = self._get_document_type()

        for key, values_list in self.params.dict_of_lists().items():
            str_values = [v for v in values_list if isinstance(v, str)]
            dtm.normalized_request_params[key].key = key
            dtm.normalized_request_params[key].values.extend(str_values)

        # includes Cookies
        dtm.parsed_request_headers = True

        # phase 1 headers
        for header_key, header_value in self.headers.items():
            dtm.request_headers[header_key] = header_value

        for cookie_key, cookie_value in self.cookies.items():
            dtm.normalized_cookies[cookie_key].key = cookie_key
            dtm.normalized_cookies[cookie_key].values.append(cookie_value)

        # multipart headers for any uploaded files
        for field_name, filename in self._get_file_info():
            dtm.multipart_headers.add(key=field_name, value=filename)

        return dtm

    def get_normalized_uri(self):
        """
        A best-effort to remove client-specific information from the path.
        Example:
        /user/123456/page/12 -> /user/{n}/page/{n}
        """
        if self._normalized_uri is not None:
            return self._normalized_uri

        self._normalized_uri = strip_uri(self.path)
        return self._normalized_uri

    def get_url_parameters(self):
        """
        Returns the url parameters in a list.
        Example
        /user/123456/page/12 -> ["123456", "12"]
        """
        if self._url_parameters is not None:
            return self._url_parameters

        self._url_parameters = find_parameters(self.path)
        return self._url_parameters

    def get_body(self, as_text=False, errors="ignore"):
        """
        Get the raw request body in either bytes or as a decoded string.
        Note that we do not use webob's Request.text here, because we do not want this to fail
        in the event of a decoding error.

        :param as_text: Boolean indicating if we should attempt to return a decoded string
        :param errors: String indicating the unicode error handling strategy, passed to decode()
        :return: The request body as either bytes or a decoded string
        """
        if not as_text:
            return self.body

        return self.body.decode(self.charset or "utf-8", errors=errors)

    def _get_http_version(self):
        """
        teamserver expects this field to be a string representing the HTTP version only.
        Using 'HTTP/1.1' is not acceptable and will cause vulnerabilities to be omitted from TS.
        """
        return self.http_version.split("/")[-1]

    def _get_document_type_from_header(self):
        """
        Returns the document type based on the content type header if present
        """
        content_type = self.content_type.lower()

        if not content_type:
            return None
        if "json" in content_type:
            return dtm_pb2.HttpRequest.JSON
        if "xml" in content_type:
            return dtm_pb2.HttpRequest.XML
        if "x-www-form-urlencoded" in content_type or "form-data" in content_type:
            return dtm_pb2.UserInput.PARAMETER_VALUE

        return dtm_pb2.HttpRequest.NORMAL

    def _get_document_type_from_body(self):
        body = self.get_body(as_text=True)

        if body.startswith("<?xml"):
            return dtm_pb2.HttpRequest.XML
        if re.search(r"^\s*[{[]", body):
            return dtm_pb2.HttpRequest.JSON

        return dtm_pb2.HttpRequest.NORMAL

    def _get_document_type(self):
        if self._document_type is not None:
            return self._document_type

        self._document_type = self._get_document_type_from_header()
        if self._document_type is None:
            self._document_type = self._get_document_type_from_body()

        return self._document_type

    def _get_file_info(self):
        """
        Get the field names and filenames of uploaded files
        :return: list of tuples of (field_name, filename)
        """
        file_info = []
        for f in self.POST.values():
            if hasattr(f, "filename") and hasattr(f, "name"):
                file_info.append((f.name, f.filename))
                logger.debug("Found uploaded file: %s", f.filename)

        return file_info



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import importlib.abc
import importlib.machinery
import importlib.util
import types
import sys
import ast


import contrast
from contrast.agent import scope
from contrast.agent.assess.adjusted_span import AdjustedSpan
from contrast.agent.assess.utils import clear_properties
from contrast.agent.request_context import RequestContext
from contrast.agent.settings import Settings
from contrast.patches.rewriter import contrast__add, contrast__append
from contrast.utils.decorators import fail_loudly
from contrast.utils.environ import test_environ
from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


def _load_module(source, module: types.ModuleType, filename):
    # For now this function should only be used to execute the original source code
    success = True

    if source is not None:
        try:
            code = compile(source, filename, "exec", dont_inherit=True)

            exec(code, module.__dict__)
        except Exception as ex:
            logger.debug(
                "WARNING: Failed to execute python module %s",
                filename,
                exc_info=ex,
            )
            success = False

    return success


class ContrastMetaPathFinder(importlib.abc.MetaPathFinder):
    @fail_loudly("Unexpected error in find_spec - will not rewrite this module")
    def find_spec(self, fullname, path, target=None):
        """
        The finder is in charge of finding a module's "spec". The spec includes import
        machinery metadata about the module - including its name, source file path, and
        the loader, among others.

        Here, we first use importlib's default machinery to get the spec for the module
        about to be imported. The problem with this spec is that it also uses the
        default loader, which isn't what we want. To get around this, we reuse some
        metadata and generate a new spec that points at our loader.

        It's possible that this is needlessly complicated. It's a first-pass
        implementation, so we can (and should) refactor this as we learn more.
        """
        default_spec = importlib.machinery.PathFinder.find_spec(fullname, path)

        if not default_spec:
            logger.debug(
                "WARNING: no spec found for module - fullname=<%s>, path=<%s>",
                fullname,
                path,
            )
            return None

        default_spec_origin = getattr(default_spec, "origin", "") or ""
        if not default_spec_origin.endswith(".py"):
            logger.debug(
                "Will not rewrite non *.py file - fullname=<%s>, path=<%s>",
                fullname,
                default_spec_origin,
            )
            return None

        return importlib.util.spec_from_file_location(
            fullname,
            default_spec.origin,
            loader=ContrastRewriteLoader(),
            submodule_search_locations=default_spec.submodule_search_locations,
        )


class ContrastRewriteLoader(importlib.abc.Loader):
    def create_module(self, _):
        """returning None uses the default behavior, which is fine"""
        return None

    def exec_module(self, module: types.ModuleType) -> None:
        """
        This method is responsible for actually doing the module `exec`-ing. We take
        control of this system and do the following:
        - read the original source file. We require pyc caching to be disabled for this
        - parse the source file into an AST
        - rewrite the AST
        - compile the AST into a code object
        - exec the code object

        Note that we add our custom add function to the module's globals. This prevents
        the need for import rewriting entirely
        """
        with scope.contrast_scope():
            original_source_code = None
            filename = module.__spec__.origin

            # May be None in some cases such as for namespace packages
            if filename is None:
                return

            try:
                with open(filename) as f:
                    original_source_code = f.read()

                tree = ast.parse(original_source_code)
            except Exception as ex:
                _load_module(original_source_code, module, filename)

                logger.debug("WARNING: failed to rewrite %s", filename, exc_info=ex)

                return

            if "contrast__add" not in module.__dict__:
                try:
                    module.__dict__["contrast__add"] = contrast__add
                    module.__dict__["contrast__append"] = contrast__append
                    ConcatRewriter().visit(tree)
                    ast.fix_missing_locations(tree)
                except Exception as ex:
                    logger.debug("WARNING: failed to rewrite %s", filename, exc_info=ex)
            else:
                logger.debug(
                    "WARNING: contrast__add is already defined in %s; will not rewrite",
                    filename,
                )

            if not _load_module(tree, module, filename):
                logger.debug("WARNING: failed to rewrite %s", filename)

                _load_module(original_source_code, module, filename)


class ConcatRewriter(ast.NodeTransformer):
    def visit_BinOp(self, binop: ast.BinOp):
        """
        If we see an "Add" binary operation, replace it with a call to our custom add
        function, which includes all necessary instrumentation.
        """
        self.visit(binop.left)
        self.visit(binop.right)

        if not isinstance(binop.op, ast.Add):
            return binop

        binop_replacement = ast.Call(
            func=ast.Name(id="contrast__add", ctx=ast.Load()),
            args=[binop.left, binop.right],
            keywords=[],
        )
        ast.copy_location(binop_replacement, binop)
        return binop_replacement

    def visit_AugAssign(self, node: ast.AugAssign):
        """
        If we see an "Append", `+=` operation, replace it with a call to our custom
        append function, which includes all necessary instrumentation.
        """
        self.visit(node.target)
        self.visit(node.value)

        if not isinstance(node.op, ast.Add):
            return node

        if isinstance(node.value, ast.Name):
            value_id = node.value.id
        elif isinstance(node.value, ast.Constant):
            value_id = node.value.value
        elif isinstance(node.value, ast.Attribute):
            value_id = node.value.value
        # TODO: PYT-2086 add else stmt

        if isinstance(node.target, ast.Name):
            target_id = node.target.id
        elif isinstance(node.target, ast.Attribute):
            target_id = node.target.value
        # TODO: PYT-2086 add else stmt

        # TODO: PYT-2086 This will never be satisfied by a node, so target id must be a string.
        #   To support the ast.Attribute type, we need to determine how to better handle Node passing.
        if not isinstance(target_id, str) or not isinstance(value_id, str):
            return node

        call_contrast_append_node = ast.Assign(
            targets=[ast.Name(id=target_id, ctx=ast.Store())],
            value=ast.Call(
                func=ast.Name(id="contrast__append", ctx=ast.Load()),
                args=[
                    # TODO: PYT-2086 Determine why we can't pass target & value nodes here
                    ast.Name(id=target_id, ctx=ast.Load()),
                    ast.Name(id=value_id, ctx=ast.Load()),
                ],
                keywords=[],
            ),
        )

        ast.copy_location(call_contrast_append_node, node)
        return call_contrast_append_node


@fail_loudly("Unexpected error concat_works", log_level="debug", return_value=False)
def concat_works():
    """
    Tests to see if concat via hooks patches works in this environment
    by running a str concat test as if we were in a request.

    Return: True if propagation happened, False otherwise.
    """
    logger.debug("Testing if concat works.")

    with scope.pop_contrast_scope():
        one = "one"
        two = "two"
        properties = contrast.STRING_TRACKER.track(one)
        properties.add_tag("UNTRUSTED", AdjustedSpan(0, len(one)))

        context = RequestContext(test_environ)
        with contrast.CS__CONTEXT_TRACKER.lifespan(context):
            res = one + two

    properties = contrast.STRING_TRACKER.get(res)
    result = properties is not None

    clear_properties()
    return result


def register():
    """
    Register our rewriter with the import system. After this call, any newly imported
    modules (from source code) will use our custom rewriter.

    Note that because this function is defined in the same module that defines our add
    replacement function, we never have to worry about rewriting the addition in the
    replacement function itself. If that were to occur, we would get an infinite
    recursion.

    Rewriter should only run in py3.10 and only in environments in which our default
    patching mechanism does not work.
    """
    settings = Settings()

    if not settings.is_rewriter_enabled:
        logger.debug("Rewriter disabled exiting setup")
        return

    if sys.version_info[:2] != (3, 10):
        # Don't register rewriter in a non py3.10 environment
        return

    if concat_works():
        # Don't register rewriter if concat already works.
        return

    sys.meta_path.insert(0, ContrastMetaPathFinder())

    settings.rewriter_enabled = True

    logger.debug("enabled AST rewriter")


def deregister():
    """
    Remove our rewriter from the import system. Modules that were loaded by our rewriter
    will remain rewritten.

    Return True if we find and deregister our machinery, False otherwise.
    """
    for i, finder in enumerate(sys.meta_path.copy()):
        if isinstance(finder, ContrastMetaPathFinder):
            sys.meta_path.pop(i)
            return True
    return False



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import socket
from contrast.utils.exceptions.contrast_service_exception import (
    ContrastServiceException,
)


class ServiceConfig(object):

    LOCALHOST_NAMES = ("127.0.0.1", "localhost")

    def __init__(self, config):
        self.socket = config.get("agent.service.socket")
        self.host = config.get("agent.service.host")
        self.port = config.get("agent.service.port")

        self._validate_config()

        self.is_unix = False
        self.socket_type = self.determine_socket_type()
        self.address = self.determine_address()

    def _validate_config(self):
        """
        This should throw an exception two scenarios:

        1. socket and tcp are configured
        2. socket is not configured and tcp is improperly configured
        """
        if self.has_socket() and self.has_tcp():
            raise ContrastServiceException(
                "Invalid service config, cannot have both socket + host/port."
            )

        if not self.has_socket() and not self.has_tcp():
            raise ContrastServiceException(
                "Invalid service config, must have host and port."
            )

    def has_socket(self):
        return bool(self.socket)

    def has_tcp(self):
        return bool(self.host and not self.host.isspace() and self.port)

    @property
    def uses_localhost(self):
        return self.host in self.LOCALHOST_NAMES

    @property
    def uses_unix_socket(self):
        return self.socket_type == socket.AF_UNIX

    @property
    def local_service_configured(self):
        return self.uses_localhost or self.uses_unix_socket

    def determine_socket_type(self):
        if self.socket:
            self.socket_type = socket.AF_UNIX
            self.is_unix = True
        else:
            self.socket_type = socket.AF_INET
        return self.socket_type

    def determine_address(self):
        if self.is_unix:
            self.address = self.socket
        else:
            self.address = (self.host, self.port)

        return self.address



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import os
import time
import threading

from pip._vendor import pkg_resources

from importlib.util import find_spec

from contrast.utils import get_installed_distributions
from contrast.utils.library_reader.utils import (
    normalize_file_name,
    get_top_level_directories,
    get_data,
    get_hash,
    CONTRAST_AGENT_DIST,
    PY_SUFFIX,
    SO_SUFFIX,
    append_files_loaded_to_activity,
)

from contrast.agent import scope
from contrast.api import Library
from contrast.api.dtm_pb2 import Activity
from contrast.utils.patch_utils import get_loaded_modules
from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


def get_active_library_names_from_pkg():
    library_names = ["contrast"]
    libraries = list(pkg_resources.working_set)
    for library in libraries:
        library_names.append(library.key)
    return library_names


class LibraryReader(object):
    def __init__(self, settings, send_sr_message_func, send_ts_message_func):
        self.settings = settings
        # TODO: PYT-2140 remove send sr msg func once we deprecate SR
        self.send_sr_message_func = send_sr_message_func
        self.send_ts_message_func = send_ts_message_func
        self.analysis_thread = None
        self.installed_distribution_keys = []

        # All packages installed in the site-packages/ directory
        self.installed_dists = get_installed_distributions()

    def start_library_analysis_thread(self, daemon=True):
        self.analysis_thread = threading.Thread(target=self._read_libraries)
        self.analysis_thread.daemon = daemon

        self.analysis_thread.start()

    def join_library_analysis_thread(self):
        if self.analysis_thread:
            self.analysis_thread.join()

    def _read_libraries(self):
        """
        Looks at every library installed in self.installed_dists, then calls search_dist
        on each dist that has a file top_level.txt, which is autogenerated by installation,
        specifying which files/directories are associated with the package

        :exception: IOError if top_level.txt, METADATA, or RECORD cannot be found
        :return: None
        """
        with scope.contrast_scope():
            logger.debug("Analyzing libraries...")

            all_dists = read_dists(self.installed_dists)
            reportable_dists = []

            for dist in all_dists:
                library_name = dist.get("file_path", "")
                if library_name and library_name != CONTRAST_AGENT_DIST:
                    reportable_dists.append(dist)

            self._send_analysis_results_appstart(reportable_dists)
            self._send_analysis_results_files_loaded(reportable_dists)

    def _send_analysis_results_appstart(self, reportable_dists):
        """
        Send library discovery in ApplicationUpdate message.
        :param reportable_dists: A list of dictionaries containing information about
            each distribution.
        """
        if not reportable_dists:
            return

        libraries = [Library(dist) for dist in reportable_dists]

        logger.debug(
            "Discovered libraries: %s",
            [lib.file_path for lib in libraries],
        )

        logger.debug("Sending ApplicationUpdate message with library analysis results")

        # Import here to prevent circular import
        from contrast.reporting import teamserver_messages

        app_update_msg = teamserver_messages.ApplicationUpdate(libraries)
        self.send_ts_message_func(app_update_msg)

    def _send_analysis_results_files_loaded(self, reportable_dists):
        if not reportable_dists:
            return

        if self.settings.config.is_service_bypassed:
            from contrast.reporting import teamserver_messages

            msg = teamserver_messages.ObservedLibraryUsage(reportable_dists)

            logger.debug(
                "Sending list of modules loaded to Contrast", direct_to_teamserver=1
            )

            self.send_ts_message_func(msg)
        else:
            total_loaded_files_cnt = 0
            activity_dtm = Activity()
            send_msg = False

            for dist in reportable_dists:
                if len(dist["used_files"]) > 0:
                    send_msg = True
                    dist_hash = dist["hash_code"]

                    total_loaded_files_cnt += append_files_loaded_to_activity(
                        activity_dtm, dist["used_files"], dist_hash
                    )

            if send_msg:
                self.send_sr_message_func([activity_dtm])

                logger.debug(
                    "Sending Activity message with %s files loaded",
                    total_loaded_files_cnt,
                )


def read_dists(installed_dists):
    results = []

    for dist in installed_dists:
        analysis_result = search_dist(dist)

        if analysis_result:
            results.append(analysis_result)

    return results


def search_dist(dist):
    """
    Searches directories related to dist, gathering relevant statistics and metadata for
    LibraryDTM. Then, assuming library was loaded, appends that metadata to the self.libraries
    in a LibraryDTM

    Created package is added to the output for the process

    :param dist: contains a pkg_resources.DistInfoDistribution object
    referencing the current library being looked for
    :return: A dictionary containing contrast-relevant information about the current distribution
    """
    # __init__.py is required in order to be a package. If there are multiple subpackages
    # we want to make sure we include all __init__.py files as part of the count
    total_files = list()
    total_used_files = set()
    top_level_dirs = get_top_level_directories(dist)

    for directory in top_level_dirs:
        package_path = os.sep.join([dist.location, directory])

        all_dir_files = get_all_files(package_path)
        dir_used_files = search_modules(package_path, directory)

        total_files.extend(all_dir_files)
        total_used_files.update(dir_used_files)

    file_count = len(total_files)
    # Set the package path to be the first directory mentioned in top_level.txt
    name = str(dist).split(" ")[0]

    version, manifest, url = get_data(dist)

    library_hash = get_hash(name, version)

    result = create_package(version, manifest, file_count, name, url, library_hash)
    result["used_files"] = total_used_files

    return result


def create_package(version, manifest, class_count, name, url, sha):
    """
    Add package with appropriate metadata to the library; reason for this is that protobuf cannot
    be pickled.

    :param version: version of package
    :param manifest: all metadata
    :param class_count: count of files
    :param name: name of the library
    :param url: homepage of package
    :param sha: sha1 hash of the name(space)version
    :return: dict of package
    """
    current_time = int(time.time() * 1000)

    return {
        "version": version,
        "manifest": manifest,
        "class_count": class_count,
        "file_path": name,
        "url": url,
        "hash_code": sha,
        "external_ms": current_time,
        "internal_ms": current_time,
    }


def search_modules(package_path, directory_name):
    """
    Searches every module to see if it is associated with the current package being looked at.
    If it is, update used file count.

    :param directory_name: name of top level directory (i.e package or subpackage name)
    :param package_path: Path to the package currently being looked for
    :return: The list of used files for this directory, package_found, name, version, manifest,
        and url
    name, version, manifest, and url are empty strings if the package was already found
    """
    used_files = set()
    sys_modules = get_loaded_modules().values()
    top_level_package = directory_name + "."

    for module in sys_modules:
        if (
            module.__name__
            and (module.__name__ + ".").startswith(top_level_package)
            and hasattr(module, "__file__")
            and module.__file__
        ):
            # First we need to check and see if the top-level package (the name up till the
            # first dot) of __name__ is the same as the current directory_name/module name

            # Then, process __file__ (Path to the loaded file)
            # __file__ is optional. If set, this attribute’s value must be a string. The import
            # system may opt to leave __file__ unset if it has no semantic meaning
            # (e.g. a module loaded from a database).

            # using os.path.realpath because in some environments there is a symlink to the
            # lib/python_version_here/site-packages directory. When checking module.__file__
            # the symlink path is used instead of the resolved path. Resolved path is needed
            # because that is what the package_path will point to.
            module_file = os.path.realpath(module.__file__)

            file_path = module_file if module_file.startswith(package_path) else ""
            if len(file_path) > 0:
                used_files.add(normalize_file_name(file_path))

    return used_files


def find_module_file(module_name, module_location):
    """
    If module is found, returns the module file path under the location specified
    by module_location.

    :param module_name: Name of the module
    :param module_location: Path to the module file
    """
    module_file = None

    try:
        # A module spec contains information used to import a module. Such information,
        # origin/location is needed here to get the location of the module found under
        # module_location. PEP 451 contains additional info on this.
        module_spec = find_spec(module_name)
        if hasattr(module_spec, "origin") and module_spec.origin:
            if module_spec.origin.startswith(module_location):
                module_file = module_spec.origin
        elif hasattr(module_spec, "location") and module_spec.location:
            if module_spec.location.startswith(module_location):
                module_file = module_spec.location
    except Exception as e:
        # Raises ModuleNotFoundError (new in 3.6+) instead of AttributeError if package is in
        # fact not a package (i.e. lacks a __path__ attribute). This is something that
        # shouldn't happen so we log it here
        logger.debug(
            "WARNING - Module %s at location %s is not a package: %s",
            module_name,
            module_location,
            e,
        )

    if module_file and os.path.isfile(module_file):
        return module_file

    return None


def get_all_files(package_path):
    """
    Get all .py and .so files in all directories in package_path.

    :param package_path: Path to the current package being looked or a path to the module file.
        If its a module, we need to get the full path to the module file
    since package_path only includes the first portion of the module name.
    :return: list of files
    """
    total_files = []

    if not os.path.isdir(package_path):
        module_name = os.path.basename(package_path)
        module_location = os.path.dirname(package_path)

        module_file = find_module_file(module_name, module_location)
        if module_file:
            total_files.append(module_file)
        # These next 2 cases can occur if __location__ or __origin__ exists as an attr
        # but they don't point to the location of the module
        # This happened for bottle
        elif os.path.exists(package_path + PY_SUFFIX):
            total_files.append(package_path + PY_SUFFIX)

        return total_files

    for _, _, files in os.walk(package_path):
        for f in files:
            if f.endswith((PY_SUFFIX, SO_SUFFIX)):
                total_files.append(f)

    return total_files



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import inspect

import contrast
from contrast.agent.settings import Settings
from contrast.api.attack import Attack, ProtectResponse
from contrast.api.user_input import UserInput
from contrast.api.sample import Sample
from contrast.agent.protect.rule import ProtectionRule
from contrast.api.settings_pb2 import InputAnalysisResult
from contrast.utils.decorators import fail_loudly, fail_quietly
from contrast.utils.decorators import cached_property
from contrast.utils.exceptions.security_exception import SecurityException
from contrast.utils.stack_trace_utils import build_and_clean_stack
from contrast.extern import structlog as logging
from contrast.utils.string_utils import ensure_string

logger = logging.getLogger("contrast")


BLOCKING_RULES = frozenset([ProtectionRule.BLOCK, ProtectionRule.BLOCK_AT_PERIMETER])
PREFILTER_RULES = frozenset([ProtectionRule.BLOCK_AT_PERIMETER])
POSTFILTER_RULES = frozenset(
    [ProtectionRule.BLOCK, ProtectionRule.MONITOR, ProtectionRule.PERMIT]
)

PREFILTER = "prefilter"
INFILTER = "infilter"
POSTFILTER = "postfilter"


class BaseRule(object):
    """
    Base rule object that all protection rules will inherit
    """

    BITMASK_VALUE = None
    RULE_NAME = "base-rule"

    def __init__(self):
        self.settings = Settings()
        self.settings.protect_rules[self.name] = self

    @property
    def name(self):
        return self.RULE_NAME

    @property
    def mode(self):
        """
        Return the mode for this rule based.

        Order of precedence:
        1. Config (contract_security.yaml)
        2. Settings from TS (Settings DTM)
        3. Default mode

        We have to assert if config_mode is not None because 0 is falsy
        """
        config_mode = self.mode_from_config()

        return config_mode if config_mode is not None else self.mode_from_settings()

    @cached_property
    def config_rule_path_mode(self):
        return f"protect.rules.{self.name}.mode"

    def mode_from_config(self):
        """
        Retrieve the mode based on the rule name

        Return None if it does no exist in the config which means default to settings
        """
        return self.settings.config.get(self.config_rule_path_mode, None)

    def mode_from_settings(self):
        """
        Retrieve the mode based on the rule name
        """
        app_settings = (
            self.settings.direct_ts_app_settings
            if self.settings.config.is_service_bypassed
            else self.settings.application_settings
        )
        for definition in app_settings.protection_rules:
            # "name" here is the "id" in the Settings DTM
            if definition.id == self.name:
                return definition.mode
        return ProtectionRule.NO_ACTION

    def is_prefilter(self):
        """
        Checks if a rules mode is for prefilter
        """
        return self.enabled and self.mode in PREFILTER_RULES

    def is_postfilter(self):
        """
        Checks if a rules mode is for postfilter
        """
        return self.enabled and self.mode in POSTFILTER_RULES

    def is_blocked(self):
        """
        Checks if a rules mode is for blocking
        """
        return self.enabled and self.mode in BLOCKING_RULES

    @property
    def enabled(self):
        """
        A rule is enabled only if all 3 conditions are met:
        1. protect is enabled in both config and server settings
        2. rule is not in disabled rules list
        3. rule mode is not NO_ACTION
        4. an exclusion wasn't applied from Teamserver
        """
        if not self.settings.is_protect_enabled():
            return False

        disabled_rules = self.settings.config.get("protect.rules.disabled_rules", [])
        if disabled_rules and self.name in disabled_rules:
            return False

        req_ctx = contrast.CS__CONTEXT_TRACKER.current()
        if req_ctx is not None and req_ctx.excluded_protect_rules:
            return self.name not in req_ctx.excluded_protect_rules

        return self.mode != ProtectionRule.NO_ACTION

    def should_block(self, attack):
        return attack and attack.response == ProtectResponse.BLOCKED

    def excluded(self, exclusions):
        """
        Check if rule is being excluded from evaluation
        :param exclusions:
        :return: True if excluded, else False
        """

        if not exclusions or len(exclusions) == 0:
            return False

        logger.debug("Checking %s exclusion(s) in %s", len(exclusions), self.name)
        return any(ex.match_protect_rule(self.name) for ex in exclusions)

    def protect_excluded_by_code(self):
        """
        Checks if analysis of code needs to be ignored based on code exceptions

        Checks stack for code exceptions
        """
        code_exclusions = self.settings.code_exclusion_matchers
        if not code_exclusions or len(code_exclusions) == 0:
            return False

        current_frame = inspect.currentframe()
        called_frame = inspect.getouterframes(current_frame, 2)
        logger.debug("caller name: %s", called_frame[1][3])
        stack = inspect.stack()

        return any(
            m.match_protect_rule(self.name) and m.match_code(stack)
            for m in code_exclusions
        )

    def prefilter(self):
        """
        Scans the input analysis for the rule and looks for matched attack signatures

        Will throw a SecurityException if a response needs to be blocked
        """
        logger.debug("PROTECT: Prefilter for %s", self.name)

        attack = self.find_attack(analysis_stage=PREFILTER)
        if attack is None or len(attack.samples) == 0:
            return

        self._append_to_context(attack)

        if attack.response == ProtectResponse.BLOCKED_AT_PERIMETER:
            raise SecurityException(
                self, "Rule triggered in prefilter. Request blocked."
            )

    def infilter(self, match_string, **kwargs):
        """
        Scans the input analysis for the rule and looks for matched attack signatures. The call to this method may be
        rule specific and include additional context in a args list.
        """
        if self.mode in [ProtectionRule.NO_ACTION, ProtectionRule.PERMIT]:
            return

        logger.debug("PROTECT: Infilter for %s", self.name)

        attack = self.find_attack(match_string, analysis_stage=INFILTER, **kwargs)
        if attack is None or len(attack.samples) == 0:
            return

        self._append_to_context(attack)

        if self.should_block(attack):
            raise SecurityException(self, f"Rule triggered. {match_string} blocked.")

    @fail_loudly("Failed to run protect rule")
    def protect(self, patch_policy, user_input, args, kwargs):
        if not self.enabled:
            return

        if self.skip_protect_analysis(user_input, args, kwargs):
            return

        self.increase_query_count()

        user_input = self.convert_input(user_input)
        if not user_input:
            return

        self.log_safely(patch_policy.method_name, user_input)

        self.infilter(user_input, **self.infilter_kwargs(user_input, patch_policy))

    def infilter_kwargs(self, user_input, patch_policy):
        return {}

    def skip_protect_analysis(self, user_input, args, kwargs):
        """
        We only want to run protect on user input that is of a type supported
        by the rule.

        Most rules use this implementation, but some override this depending on
        expected user input types.

        :return: Bool if to skip running protect infilter
        """
        if not user_input:
            return True

        if isinstance(user_input, (str, bytes)):
            return False

        logger.debug(
            "WARNING: unknown input type %s for rule %s", type(user_input), self.name
        )

        return True

    def convert_input(self, user_input):
        return ensure_string(user_input)

    def increase_query_count(self):
        """Only rules for database support increase the query count"""
        pass

    def postfilter(self):
        """
        Scans the input analysis for the rule and looks for matched attack signatures

        Appends attacker to the context if a positive evaluation is found
        """
        logger.debug("PROTECT: Postfilter for %s", self.name)

        if self.mode in [ProtectionRule.NO_ACTION, ProtectionRule.PERMIT]:
            return

        attack = self.find_attack(analysis_state=POSTFILTER)
        if attack is None or len(attack.samples) == 0:
            return

        self._append_to_context(attack)

        if self.should_block(attack):
            raise SecurityException(
                self, f"Rule triggered in postfilter. {self.name} blocked."
            )

    def find_attack(self, candidate_string=None, analysis_stage=None, **kwargs):
        """
        Finds the attacker in the original string if present
        """
        if candidate_string is not None:
            logger.debug("Checking for %s in %s", self.name, candidate_string)

        if self.protect_excluded_by_code():
            return None

        # if rule mode is BAP, only speedracer determines if we should block
        if (
            analysis_stage == PREFILTER
            and self.mode == ProtectionRule.BLOCK_AT_PERIMETER
        ):
            return None

        evaluations_for_rule = self.evaluations_for_rule()

        attack = None
        for evaluation in evaluations_for_rule:
            if analysis_stage == POSTFILTER and (
                evaluation.attack_count > 0
                or evaluation.input_type == UserInput.QUERYSTRING
            ):
                continue

            if candidate_string:
                if candidate_string.find(evaluation.value) == -1:
                    continue

                attack = self.build_attack_with_match(
                    candidate_string, evaluation, attack, **kwargs
                )
            else:
                attack = self.build_attack_without_match(evaluation, attack, **kwargs)

        return attack

    def build_attack_with_match(
        self, candidate_string, evaluation=None, attack=None, **kwargs
    ):
        attack = self.build_or_append_attack(
            evaluation, attack, candidate_string, **kwargs
        )

        if evaluation:
            evaluation.attack_count += 1

        attack.set_response(self.response_from_mode(self.mode))
        self.log_rule_matched(evaluation, attack.response, candidate_string)
        return attack

    def build_attack_without_match(self, evaluation=None, attack=None, **kwargs):
        if self.mode == ProtectionRule.BLOCK_AT_PERIMETER:
            attack = self.build_or_append_attack(evaluation, attack, **kwargs)

            attack.set_response(self.response_from_mode(self.mode))
            self.log_rule_matched(evaluation, attack.response)
        elif evaluation is None or (
            evaluation.attack_count == 0
            and evaluation.score_level > InputAnalysisResult.WORTHWATCHING
        ):
            # an exploit was found first so we do not need to probe this
            attack = self.build_or_append_attack(evaluation, attack, **kwargs)
            attack.set_response(ProtectResponse.PROBED)
            self.log_rule_probed(evaluation)

        return attack

    def build_or_append_attack(
        self, evaluation, attack=None, candidate_string=None, **kwargs
    ):
        if attack is None:
            attack = self.build_base_attack()

        attack.add_sample(self.build_sample(evaluation, candidate_string, **kwargs))

        return attack

    def build_base_attack(self):
        return Attack(self.name)

    def build_sample(self, evaluation, candidate_string, **kwargs):
        return self.build_base_sample(evaluation)

    def build_user_input(self, evaluation):
        return UserInput(
            input_type=evaluation.input_type,
            key=evaluation.key,
            value=evaluation.value,
            path=evaluation.path,
            matcher_ids=[idx for idx in evaluation.ids],
        )

    def build_base_sample(self, evaluation, prebuilt_stack=None):
        sample = Sample()

        if evaluation:
            sample.set_user_input(self.build_user_input(evaluation))
        else:
            sample.set_user_input(
                UserInput(
                    input_type=0,
                    key="",
                    value="",
                )
            )

        stack = prebuilt_stack if prebuilt_stack else build_and_clean_stack()
        sample.set_stack(stack)

        return sample

    def log_rule_matched(self, evaluation, response, _=None):
        """
        Logs the exploit for the rule to the security logger
        """
        if evaluation:
            key = str(evaluation.key) if evaluation.key else ""
            input_type_name = UserInput.type_to_str(evaluation.input_type)

            rule_message = (
                f"{self.name} - {ensure_string(evaluation.value, errors='replace')}"
            )

            message = f"The {input_type_name} {key} had a value that successfully exploited: {ensure_string(rule_message, errors='replace')}"
        else:
            message = self.effective_attack_message()

        logger.debug("%s %s", ProtectResponse.to_str(response), message)

    def log_rule_probed(self, evaluation):
        """
        Logs the probed attack for the rule to the security logger
        """

        if evaluation:
            key = str(evaluation.key) if evaluation.key else ""

            input_type_name = UserInput.type_to_str(evaluation.input_type)

            rule_message = (
                f"{self.name} - {ensure_string(evaluation.value, errors='replace')}"
            )

            message = (
                f"The {input_type_name} {key} had a value that matched a signature but did not"
                f" exploit: {ensure_string(rule_message, errors='replace')}"
            )
        else:
            message = self.effective_attack_message()

        logger.debug(message)

    def effective_attack_message(self):
        return f"An effective attack was detected against {self.name}."

    def _append_to_context(self, attack):
        """
        Appends the current context's request dtm to the defend activity along with attacks to the rules
        """
        context = contrast.CS__CONTEXT_TRACKER.current()
        if context is None:
            # do not remove; this case is not yet well-understood
            logger.debug("WARNING: failed to get request context in _append_to_context")
            return

        context.attacks.append(attack)

    _RESPONSE_MAP = {
        ProtectionRule.MONITOR: ProtectResponse.MONITORED,
        ProtectionRule.BLOCK: ProtectResponse.BLOCKED,
        ProtectionRule.BLOCK_AT_PERIMETER: ProtectResponse.BLOCKED_AT_PERIMETER,
        ProtectionRule.NO_ACTION: ProtectResponse.NO_ACTION,
        ProtectionRule.PERMIT: ProtectResponse.NO_ACTION,
    }

    def response_from_mode(self, mode):
        return self._RESPONSE_MAP.get(mode)

    def evaluations_for_rule(self):
        context = contrast.CS__CONTEXT_TRACKER.current()
        if context is None:
            # do not remove; this case is not yet well-understood
            logger.debug(
                "WARNING: failed to get request context in evaluations_for_rule"
            )
            return []

        if context.speedracer_input_analysis is None:
            return []

        evaluations = context.speedracer_input_analysis.results
        return [
            evaluation for evaluation in evaluations if evaluation.rule_id == self.name
        ]

    @fail_quietly("Failed to log user input for protect rule")
    def log_safely(self, method_name, user_input):
        """
        Attempt to log user supplied input but do not fail if unable to do so.
        """
        logger.debug(
            "Applying %s rule method %s with user input %s",
            self.name,
            method_name,
            ensure_string(user_input, errors="replace"),
        )



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.


class ConnectionStatus(object):
    """
    Stores the connection information between the agent and the service
    """

    def __init__(self):
        self._connected = False

        self._success = 0
        self._failed = 0

    @property
    def failure_count(self):
        return self._failed

    @property
    def success_count(self):
        return self._success

    @property
    def connected(self):
        return self._connected

    @property
    def resend_startup(self):
        return bool(self._success) and not self.connected

    def success(self):
        self._connected = True
        self._success += 1

    def failure(self):
        self._connected = False
        self._failed += 1



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import contrast
from contrast.agent import agent_lib
from contrast.agent.protect.rule.base_rule import BaseRule, UserInput

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


class MethodTampering(BaseRule):
    RULE_NAME = "method-tampering"
    BITMASK_VALUE = agent_lib.CONSTANTS.RuleType.get(RULE_NAME)
    USER_INPUT_KEY = UserInput.type_to_str(UserInput.METHOD)

    def postfilter(self):
        """
        At postfilter we generate activity if input analysis was found and depending on application response code.

        if response code is either 4xx or 5xx, application was not exploited (only probed) by an unexpected HTTP method.
        If response code is anything else, then an unexpected HTTP method successfully exploited the application.
        """
        logger.debug("PROTECT: Postfilter for %s", self.name)

        evaluations_for_rule = self.evaluations_for_rule()

        context = contrast.CS__CONTEXT_TRACKER.current()
        if context is None or not hasattr(context, "response"):
            # do not remove; this case is not yet well-understood
            logger.debug("WARNING: failed to get context in MethodTampering.postfilter")
            return

        response_code = context.response.status_code

        for evaluation in evaluations_for_rule:
            if str(response_code).startswith("4") or str(response_code).startswith("5"):
                attack = self.build_attack_without_match(
                    method=evaluation.value,
                    response_code=response_code,
                )
            else:
                attack = self.build_attack_with_match(
                    None,
                    method=evaluation.value,
                    response_code=response_code,
                )
            self._append_to_context(attack)

    def build_sample(self, evaluation, candidate_string, **kwargs):
        sample = self.build_base_sample(None)

        method = kwargs.get("method", "")

        sample.details["method"] = method
        sample.details["responseCode"] = kwargs.get("response_code", -1)
        sample.set_user_input(self.build_user_input(method))

        return sample

    def build_user_input(self, method):
        return UserInput(
            input_type=UserInput.METHOD,
            key=self.USER_INPUT_KEY,
            value=method,
        )



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Functions for communication with speedracer to receive input analysis
"""
import contrast
from contrast.agent import service_client
from contrast.agent.settings import Settings
from contrast.agent.protect.rule import ProtectionRule
from contrast.utils.exceptions.security_exception import SecurityException

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


def get_input_analysis():
    logger.debug("Getting input analysis from speedracer ...")

    context = contrast.CS__CONTEXT_TRACKER.current()
    message = context.activity.http_request

    responses = service_client.send_messages_get_responses([message])
    speedracer_response = responses[0] if responses else None

    if not speedracer_response or not speedracer_response.protect_state:
        logger.debug("No response from speedracer: %s", speedracer_response)
        return None

    if speedracer_response.input_analysis is None:
        logger.debug(
            "Speedracer returned nil input analysis - no evaluation for request"
        )

    context.do_not_track = speedracer_response.protect_state.track_request

    if speedracer_response.protect_state.security_exception:
        create_activity_from_speedracer_input_analysis(
            speedracer_response.input_analysis, context
        )
        raise SecurityException(context, "Speedracer said to block this request")

    logger.debug("Speedracer input analysis: %s", speedracer_response.input_analysis)
    return speedracer_response.input_analysis


def create_activity_from_speedracer_input_analysis(speedracer_input_analysis, context):
    """
    If speedracer returns any input analysis results, we should create
    attack result samples in case it did not create it.

    """
    if speedracer_input_analysis is None:
        return

    settings = Settings()

    for evaluation in speedracer_input_analysis.results:
        rule = settings.protect_rules[evaluation.rule_id]
        if rule.mode == ProtectionRule.BLOCK:
            # special case for rules (xss) that used to have infilter but now are only prefilter / BAP
            attack = rule.build_attack_with_match(evaluation.value, evaluation)
        else:
            attack = rule.build_attack_without_match(evaluation=evaluation)
        context.attacks.append(attack)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
class ProtectionRule:
    NO_ACTION = 0
    MONITOR = 1
    BLOCK = 2
    BLOCK_AT_PERIMETER = 3
    PERMIT = 4



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import os
import socket
import struct

import contrast
from contrast.agent.connection_status import ConnectionStatus
from contrast.agent.service_config import ServiceConfig
from contrast.agent.settings import Settings
from contrast.api.dtm_pb2 import (
    Activity,
    AgentStartup,
    ApplicationCreate,
    HttpRequest,
    Message,
    Poll,
)
from contrast.api.settings_pb2 import AgentSettings
from contrast.utils.exceptions.contrast_service_exception import (
    ContrastServiceException,
)
from contrast.utils.singleton import Singleton
from contrast.utils import service_util
from contrast.utils.service_util import ServiceUtil
from contrast.utils.timer import now_ms

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


_RECV_FMT = ">I"
_APPLICATION_LANGUAGE = "Python"


class _ServiceState(Singleton):
    def init(self):
        """
        Singletons should override init, not __init__.
        """
        self.count = 0
        self.connection_status = ConnectionStatus()
        self.bundled_flag = "external"
        self.service_config = ServiceConfig(Settings().config)
        self.poll = Poll()

    @property
    def local_service_configured(self):
        return self.service_config.local_service_configured

    @property
    def client_id(self):
        """
        Client id is used by Speedracer to recognize when messages are sent by different
        worker processes of the same application.
        client_id must be the parent process id as it is the same for all workers serving an app.
        :return: str app name and parent process id
        """
        return str(Settings().app_name) + "-" + str(os.getppid())

    @property
    def app_path(self):
        """
        Return the path for where the application is running, either configured
        in common config or by looking at the current working dir.
        """
        return (
            Settings().config.get("application.path") or contrast.AGENT_CURR_WORKING_DIR
        )

    @property
    def resend_startup(self):
        return self.connection_status.resend_startup

    def start_service(self):
        """
        Start the bundled service.

        Should not be run if the user would like to connect to an external service.

        :return: bool if bundled Contrast Service was started.
        """
        settings = Settings()

        logger.info(
            "Attempted to start bundled Contrast Service for application %s.",
            settings.app_name,
        )
        logger.info(
            "If socket already exists at %s, will attempt to connect to that instead",
            self.service_config.address,
        )

        path = settings.config.get("agent.service.logger.path", "contrast_service.log")

        if not ServiceUtil(path).start_bundled_service():
            logger.error("Contrast Service did not start")
            return False

        return True

    def initialize_and_send_messages(self):
        """
        Initializing the connection to Contrast Service requires checking if the app is configured
        to use the bundled service and turning it on if so.

        If the app isn't configured to use the bundled service or after turning the bundled service on,
        agent sends initial messages to Contrast Service (AgentStartup, ApplicationCreate)
        to determine that it is successfully able to communicate with it.

        If configuration specifies an external service, whatever manages the external service should handle cases
        of the service being down. The agent should not be responsible for managing service state in this case.

        :return: list of Message responses

        """
        settings = Settings()

        if settings.is_bundled_enabled and self.local_service_configured:
            self.bundled_flag = "bundled"

            if not self.start_service():
                # if bundled service is configured but it could not be started,
                # then we don't have service data to process, so we should not
                # continue on with Agent startup.
                raise ContrastServiceException

        logger.info("Will communicate with Contrast Service")
        self._build_and_send_startup_msgs()

    def _build_and_send_startup_msgs(self):
        settings = Settings()

        first_msg = self._build_agent_startup_msg(settings)

        # it's only necessasry to wait to send messages when using the bundled service
        # because it takes a bit to start up
        wait = 1 if self.bundled_flag == "bundled" else 0

        responses = self.send_messages_retry([first_msg], wait=wait)
        if not responses:
            # If we received no response for the first message, we should not
            # try to send the second message.
            raise ContrastServiceException

        settings.process_responses(responses)

        # now that we have server features from first_msg we can build + send app create
        second_msg = self._build_app_create_msg(settings)

        # no wait because we know SR is already up
        responses = self.send_messages_retry([second_msg])
        if not responses:
            # If we received no response for the second message, we should not
            # continue on with Agent startup.
            raise ContrastServiceException

        settings.process_responses(responses)

    def send_messages_retry(self, messages, attempts=0, wait=0, raise_exception=True):
        """
        ****************
        When using a nonzero value for `wait`, this function will cause the main python
        process to sleep for an exponentially increasing amount of time on each retry
        attempt.
        ****************

        Attempt to send a list of messages to Speedracer. If sending messages fails, this method
        recursively calls itself to attempt to resend 3 times.

        If after 3 times sending messages still failed, the connection status is set to faise
        and the exception may or may not be raised.

        :param messages: message(s) to to send to Speedracer
        :param attempts: a count of number of attempts to send to messages.
        :param wait: seconds to wait before sending the messages.
            This is used when the bundled service is getting started (start_service has been called)
            because sometimes it takes a second for it to get setup.
        :param raise_exception: bool to determine if to raise an exception or not
            The caller of this method will determine if to ask this method to raise the exception or not.
            Currently the logic is that the caller will ask to raise the exception if the caller itself is catching for it.

            Current callers that will ask for this exception to be raised are:
            1. when initializing communication to Speedracer, because the middleware needs to know if this initialization
                has failed to set settings initialization to false.
            2. when input analysis messages are send to Speedracer, because the middleware needs to know if
                this analysis has failed to re-route the request to call_without_agent.

            Current callers that will not want the exception to be raised:
            1. when final messages are sent to Speedracer for reporting. The middleware is not equipped
                to handle an exception at that time, so if sending these messages fails, the reporting will fail
                but the application should not.

        :return: list of responses from Speedracer
        :side effect: ContrastServiceException if raise_exception is True
        """
        if wait > 0:
            logger.info(
                "Waiting %s second%s for Contrast Service",
                wait,
                "" if wait == 1 else "s",
            )
            service_util.sleep(wait)

        try:
            return self._send_messages(messages)
        except ContrastServiceException:
            logger.debug(
                "Trying to send messages, but received ContrastServiceException"
            )

            if attempts < 3:
                new_attempts = attempts + 1
                new_wait = wait * 2
                logger.debug(
                    "Retrying to send messages: retry attempt #%s, wait=%s",
                    new_attempts,
                    new_wait,
                )
                return self.send_messages_retry(
                    messages, new_attempts, new_wait, raise_exception
                )

            logger.warning(
                "Unable to send messages to Contrast Service after 3 retries."
            )
            self.connection_status.failure()
            if raise_exception:
                raise

        return []

    def send_dtm(self, message):
        """
        This takes in a json string so that we can encode and decode in separate tasks

        :param message: json string representation of the dtm object
        :return: json string of the response
        :side effect: ContrastServiceException if any problems connecting / sending messages occur
        """
        packed = message.SerializeToString()
        sock = None
        try:
            sock = socket.socket(self.service_config.socket_type, socket.SOCK_STREAM)
            sock.connect(self.service_config.address)
            self.send_message(sock, packed)
            response = self.receive_message(sock)

            if response is not None:
                settings = AgentSettings()
                settings.ParseFromString(response)
                return settings

            logger.warning("Received an empty response from Contrast Service")
        except Exception:
            # If there are any issues connecting or sending messages to Contrast Service,
            # we raise a special exception for middleware handle_exception to catch but not raise
            # so the request can be processed without the agent.
            raise ContrastServiceException
        finally:
            if sock is not None:
                sock.close()

        return None

    def _send_messages(self, messages):
        responses = []
        messages = [messages] if not isinstance(messages, list) else messages

        to_send = self.update_messages(messages)

        logger.debug("Sending %s messages to Contrast Service", len(to_send))
        for agent_message, msg in zip(messages, to_send):
            logger.debug(
                "Sending %s message with msg_count=%s, pid=%s",
                agent_message.__class__.__name__,
                msg.message_count,
                msg.pid,
            )

            response = self.send_dtm(msg)
            if response is not None:
                logger.debug(
                    "Received response for message with msg_count=%s, pid=%s",
                    msg.message_count,
                    msg.pid,
                )
                responses.append(response)
            else:
                logger.debug(
                    "No response for message with msg_count=%s, pid=%s",
                    msg.message_count,
                    msg.pid,
                )

        self.connection_status.success()
        return responses

    def send_message(self, sock, msg):
        """
        Prefix each message with a 4-byte length (network byte order)
        """
        msg = struct.pack(_RECV_FMT, len(msg)) + msg
        sock.sendall(msg)

    def receive_message(self, sock):
        """
        Read message length and unpack it into an integer

        :param sock - socket to receive information from
        :return bytes if successful else None
        """
        raw_message_length = self.receive_all(sock, 4)

        if not raw_message_length:
            logger.debug("Falsy raw message length of %s", raw_message_length)
            return None

        message_length = struct.unpack(_RECV_FMT, raw_message_length)[0]
        return self.receive_all(sock, message_length)

    def receive_all(self, sock, n):
        """
        Helper function to recv n bytes or return None if EOF is hit
        """
        buffer_array = bytearray()
        while len(buffer_array) < n:
            packet = sock.recv(n - len(buffer_array))
            if not packet:
                return None
            buffer_array.extend(packet)
        return bytes(buffer_array)

    def build_empty_message(self):
        settings = Settings()

        message = Message()
        message.app_name = settings.app_name
        message.app_path = self.app_path
        message.app_language = _APPLICATION_LANGUAGE
        message.client_id = self.client_id
        message.client_number = 1
        message.client_total = 1
        message.timestamp_ms = now_ms()
        message.pid = settings.pid

        self.count += 1
        message.message_count = self.count

        return message

    def update_message(self, agent_message):
        message = self.build_empty_message()

        if isinstance(agent_message, Activity):
            agent_message.duration_ms = now_ms()
            message.activity.CopyFrom(agent_message)

        if isinstance(agent_message, AgentStartup):
            message.agent_startup.CopyFrom(agent_message)

        if isinstance(agent_message, ApplicationCreate):
            message.application_create.CopyFrom(agent_message)

        if isinstance(agent_message, HttpRequest):
            message.prefilter.CopyFrom(agent_message)

        if isinstance(agent_message, Poll):
            message.poll.CopyFrom(agent_message)

        return message

    def update_messages(self, messages):
        """
        Possible messages supported currently

        contrast.api.dtm.AgentStartup agent_startup = 11;
        contrast.api.dtm.ApplicationCreate application_create = 12;
        contrast.api.dtm.Activity activity = 14;
        contrast.api.dtm.Poll
        """
        updated = []
        for item in messages:
            prepared_msg = self.update_message(item)
            updated.append(prepared_msg)

        return updated

    def _build_agent_startup_msg(self, settings):
        agent_startup = AgentStartup()

        agent_startup.server_name = settings.get_server_name()
        agent_startup.server_path = settings.get_server_path()
        agent_startup.server_type = settings.get_server_type()
        agent_startup.server_version = contrast.__version__

        agent_startup.environment = settings.config.get("server.environment")

        agent_startup.server_tags = settings.config.get("server.tags")
        agent_startup.application_tags = settings.config.get("application.tags")
        agent_startup.library_tags = settings.config.get("inventory.tags")
        agent_startup.finding_tags = settings.config.get("assess.tags")

        agent_startup.version = contrast.__version__

        settings = Settings()

        log_info = {
            "Server Name": agent_startup.server_name,
            "Server Path": agent_startup.server_path,
            "Server Type": agent_startup.server_type,
            "Application Name": settings.app_name,
            "Application Path": self.app_path,
            "Application Language": _APPLICATION_LANGUAGE,
        }

        logger.info("Application context", **log_info)

        return agent_startup

    def _build_app_create_msg(self, settings):
        create = ApplicationCreate()
        create.group = settings.config.get("application.group", "")
        create.app_version = settings.config.get("application.version", "")
        create.code = settings.config.get("application.code", "")

        create.metadata = settings.config.get("application.metadata", "")

        # build based views
        create.session_id = settings.config.session_id
        create.session_metadata = settings.config.get_session_metadata()

        # instrumentation mode
        create.mode.protect = settings.is_protect_enabled()
        create.mode.assess = settings.is_assess_enabled()

        return create


def is_connected():
    return _ServiceState().connection_status.connected


def send_messages_get_responses(messages):
    """Send messages to service and return responses"""
    service_state = _ServiceState()

    if service_state.resend_startup:
        logger.debug("Resending startup messages")
        service_state._build_and_send_startup_msgs()

    return service_state.send_messages_retry(messages, raise_exception=False)


def send_messages(messages):
    """
    Rely on ServiceClient having already initialized its connection to Speedracer
    and ask it to send a list of messages.
    Process the responses from sending the messages.
    :param messages: a list of protobuf messages to send
    :side effect: ContrastServiceException could be raised but should be catch by the middleware
    """
    responses = send_messages_get_responses(messages)
    Settings().process_responses(responses)


def send_heartbeat_message():
    """
    Agent sends a Poll message to Speedracer as a way to maintain a heartbeat.
    Note that a Poll message should never be the first message sent to Speedracer,
    a Poll message should only be sent AFTER an AgentStartup.
    """
    return send_messages([_ServiceState().poll])


def send_startup_messages():
    """
    Initialize ServiceClient and its connection to Speedracer by asking it to send the agent and app
    initialization messages. This is successful if responses are received to process for new settings.

    :return: bool if successfuly able to send initial messages and receive and process responses.
    """
    _ServiceState().initialize_and_send_messages()



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Instrumentation for exec/eval.

This was defined outside of policy.json because we need to pass globals/locals from the
frame in which function was originally called
"""
from sys import _getframe as getframe

from contrast.extern.wrapt import register_post_import_hook

import builtins
import contrast
from contrast.agent import scope
from contrast.agent.policy.loader import Policy
from contrast.agent.assess.policy import trigger_policy
from contrast.utils.patch_utils import build_and_apply_patch
from contrast.applies.assess.unsafe_code_execution import (
    apply_rule as apply_unsafe_code_exec_rule,
)
from contrast.utils.decorators import fail_quietly


INSTRUMENTED_FRAME_DEPTH = 1


@fail_quietly("Error applying rule for exec/eval patch")
def apply_rule(rule_applicator, orig_func, result, args, kwargs):
    context = contrast.CS__CONTEXT_TRACKER.current()
    if not (context and context.propagate_assess):
        return

    with scope.contrast_scope():
        rule_applicator("BUILTIN", orig_func.__name__, result, args, kwargs)


@fail_quietly("Error running path traversal assess rule")
def apply_pt_rule(module, method, result, args, kwargs):
    policy = Policy()
    trigger_rule = policy.triggers["path-traversal"]

    trigger_nodes = trigger_rule.find_trigger_nodes(module, method)

    trigger_policy.apply(trigger_rule, trigger_nodes, result, args, kwargs)


def build_exec_eval_patch(orig_func, _, rule_applicator):
    def exec_eval_patch(code, globs=None, locs=None):
        """
        Run exec/eval call with proper context to adjust for current frame

        Code ported from six module
        See https://github.com/benjaminp/six/blob/master/six.py#L694

        Reapplying the context from the 3rd frame (from top of stack) is necessary
        because the globals and locals in that frame are used in the original call to
        exec/eval. The exception to this is if the caller passes custom globals/locals
        to the function.

        If we fail provide this context we will see a number of errors due to things
        not defined in the scope of this function upon calling the original function
        definition.
        """
        result = None

        if globs is None:
            frame = getframe(INSTRUMENTED_FRAME_DEPTH)

            globs = frame.f_globals
            if locs is None:
                locs = frame.f_locals
            del frame
        elif locs is None:
            locs = globs

        try:
            result = orig_func(code, globs, locs)
        except Exception:
            result = None
            raise
        finally:
            apply_rule(rule_applicator, orig_func, result, (code,), {})

        return result

    return exec_eval_patch


def patch_exec_and_eval(builtins_module):
    build_and_apply_patch(
        builtins_module, "eval", build_exec_eval_patch, (apply_unsafe_code_exec_rule,)
    )

    build_and_apply_patch(
        builtins_module,
        "exec",
        build_exec_eval_patch,
        (apply_unsafe_code_exec_rule,),
    )


def register_patches():
    register_post_import_hook(patch_exec_and_eval, builtins.__name__)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import contrast
from contrast.api.dtm_pb2 import HttpRequest
from contrast.api.user_input import UserInput
from contrast.utils.decorators import fail_quietly
from contrast.utils.string_utils import ensure_string
from contrast.utils.timer import now_ms
from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")

class ProtectResponse:
    NO_ACTION = 0
    BLOCKED = 1
    MONITORED = 2
    PROBED = 3
    BLOCKED_AT_PERIMETER = 4
    SUSPICIOUS = 6

    _RESPONSE_TO_STR = {
        MONITORED: "MONITORED",
        BLOCKED: "BLOCKED",
        BLOCKED_AT_PERIMETER: "BLOCKED AT PERIMETER",
        NO_ACTION: "NO ACTION",
        SUSPICIOUS: "SUSPICIOUS",
    }
    _RESPONSE_TO_TS_NODE = {
        MONITORED: "exploited",
        BLOCKED: "blocked",
        PROBED: "ineffective",
        SUSPICIOUS: "suspicious",
    }
    @classmethod
    def to_str(cls, response: int) -> str:
        """
        Convert int to str response
        """
        return cls._RESPONSE_TO_STR.get(response, "")

    @classmethod
    def to_ts_node(cls, response: int) -> str:
        """
        Convert int to TS-approved node str
        """
        return cls._RESPONSE_TO_TS_NODE.get(response)





class Attack:
    """
    Class storing all data necessary to report a protect attack.
    """
    def __init__(self, rule_id):
        self.rule_id = rule_id
        self.samples = []
        self.response = None
        self.start_time_ms = contrast.CS__CONTEXT_TRACKER.current().request.timestamp_ms
        self.response = None

    @property
    def blocked(self):
        return self.response == ProtectResponse.BLOCKED

    def add_sample(self, sample):
        self.samples.append(sample)

    def set_response(self, response):
        self.response = response

    def _convert_samples(self, request_dtm: HttpRequest):
        DOCUMENT_TYPES = {
            0: "NORMAL",
            1: "JSON",
            2: "XML",
        }

        path, _, querystring = request_dtm.raw.partition("?")
        return [
            {
                "blocked": self.blocked,
                "input": {
                    "documentPath": sample.user_input.path,
                    "documentType": DOCUMENT_TYPES.get(sample.user_input.document_type),
                    "filters": sample.user_input.matcher_ids,
                    "name": sample.user_input.key,
                    "time": sample.timestamp_ms,
                    "type": UserInput.type_to_str(sample.user_input.input_type),
                    "value": sample.user_input.value,
                },
                "details": sample.details,
                "request": {
                    "body": ensure_string(request_dtm.request_body_binary),
                    # the WSGI environ supports only one value per request header. However
                    # the server decides to handle multiple headers, we're guaranteed to
                    # have only unique keys in request.request_headers (since we iterate
                    # over webob's EnvironHeaders). Thus, each value list here is length-1.
                    "headers": {
                        k: [v] for k, v in request_dtm.request_headers.items()
                    },
                    "method": request_dtm.method,
                    "parameters": {
                        h.key: list(h.values)
                        for h in request_dtm.normalized_request_params.values()
                    },
                    "port": request_dtm.receiver.port,
                    "protocol": request_dtm.protocol,
                    "queryString": querystring,
                    "uri": path,
                    "version": request_dtm.version,
                },
                "stack": self._convert_stacks(sample.stack_trace_elements),
                "timestamp": {
                    "start": sample.timestamp_ms,  # in ms
                    "elapsed": (now_ms() - sample.timestamp_ms),  # in ms which is the format TS accepts
                },
            }
            for sample in self.samples
        ]

    def _convert_stacks(self, stacks_dtm):
        return [
            {
                "declaringClass": stack.declaring_class,
                "methodName": stack.method_name,
                "fileName": stack.file_name,
                "lineNumber": stack.line_number,
            }
            for stack in stacks_dtm
        ]

    @fail_quietly("Unable to create time map", return_value={})
    def _create_time_map(self, samples):
        """
        For the list of samples, createa dict of:

        second since attack start => attacks in that second.
        """
        time_map = {}

        for sample in samples:
            elapsed_secs = sample["timestamp"]["elapsed"]
            time_map.setdefault(elapsed_secs, 0)
            time_map[elapsed_secs] += 1

        return time_map

    def to_json(self, request_dtm: HttpRequest):
        common_fields = {
            "startTime": 0,
            "total": 0,
        }
        json = {
            "startTime": self.start_time_ms,
            "blocked": common_fields,
            "exploited": common_fields,
            "ineffective": common_fields,
            "suspicious": common_fields,
        }

        relevant_mode = ProtectResponse.to_ts_node(self.response)
        if relevant_mode is None:
            # Don't know what response is so just report default info so we can debug.
            return json

        samples = self._convert_samples(request_dtm)

        json[relevant_mode] = {
            "total": 1,  # always 1 until batching happens
            "startTime": self.start_time_ms,
            "attackTimeMap": self._create_time_map(samples),
            "samples": samples,
        }

        return json



# SPDX-License-Identifier: MIT OR Apache-2.0
# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the MIT License.  See the LICENSE file in the root of this
# repository for complete details.

"""
Type information used throughout ``structlog``.

For now, they are considered provisional. Especially `BindableLogger` will
probably change to something more elegant.

.. versionadded:: 20.2
"""

import sys

from types import TracebackType
from typing import (
    Any,
    Callable,
    Dict,
    Mapping,
    MutableMapping,
    Optional,
    TextIO,
    Tuple,
    Type,
    Union,
)


# This construct works better with Mypy.
# Doing the obvious ImportError route leads to an 'Incompatible import of
# "Protocol"' error.
if sys.version_info >= (3, 8):
    from typing import Protocol, runtime_checkable
else:
    from contrast.extern.typing_extensions import Protocol, runtime_checkable


WrappedLogger = Any
"""
A logger that is wrapped by a bound logger and is ultimately responsible for
the output of the log entries.

``structlog`` makes *no* assumptions about it.

.. versionadded:: 20.2
"""


Context = Union[Dict[str, Any], Dict[Any, Any]]
"""
A dict-like context carrier.

.. versionadded:: 20.2
"""


EventDict = MutableMapping[str, Any]
"""
An event dictionary as it is passed into processors.

It's created by copying the configured `Context` but doesn't need to support
copy itself.

.. versionadded:: 20.2
"""

Processor = Callable[
    [WrappedLogger, str, EventDict],
    Union[Mapping[str, Any], str, bytes, bytearray, Tuple[Any, ...]],
]
"""
A callable that is part of the processor chain.

See :doc:`processors`.

.. versionadded:: 20.2
"""

ExcInfo = Tuple[Type[BaseException], BaseException, Optional[TracebackType]]
"""
An exception info tuple as returned by `sys.exc_info`.

.. versionadded:: 20.2
"""


ExceptionFormatter = Callable[[TextIO, ExcInfo], None]
"""
A callable that pretty-prints an `ExcInfo` into a file-like object.

Used by `structlog.dev.ConsoleRenderer`.

.. versionadded:: 21.2
"""


@runtime_checkable
class BindableLogger(Protocol):
    """
    **Protocol**: Methods shared among all bound loggers and that are relied on
    by ``structlog``.

    .. versionadded:: 20.2
    """

    _context: Context

    def bind(self, **new_values: Any) -> "BindableLogger":
        ...

    def unbind(self, *keys: str) -> "BindableLogger":
        ...

    def try_unbind(self, *keys: str) -> "BindableLogger":
        ...

    def new(self, **new_values: Any) -> "BindableLogger":
        ...


class FilteringBoundLogger(BindableLogger, Protocol):
    """
    **Protocol**: A `BindableLogger` that filters by a level.

    The only way to instantiate one is using `make_filtering_bound_logger`.

    .. versionadded:: 20.2.0
    """

    def debug(self, event: str, **kw: Any) -> Any:
        """
        Log *event* with **kw** at **debug** level.
        """

    def info(self, event: str, **kw: Any) -> Any:
        """
        Log *event* with **kw** at **info** level.
        """

    def warning(self, event: str, **kw: Any) -> Any:
        """
        Log *event* with **kw** at **warn** level.
        """

    def warn(self, event: str, **kw: Any) -> Any:
        """
        Log *event* with **kw** at **warn** level.
        """

    def error(self, event: str, **kw: Any) -> Any:
        """
        Log *event* with **kw** at **error** level.
        """

    def err(self, event: str, **kw: Any) -> Any:
        """
        Log *event* with **kw** at **error** level.
        """

    def fatal(self, event: str, **kw: Any) -> Any:
        """
        Log *event* with **kw** at **critical** level.
        """

    def exception(self, event: str, **kw: Any) -> Any:
        """
        Log *event* with **kw** at **error** level and ensure that ``exc_info``
        is set in the event dictionary.
        """

    def critical(self, event: str, **kw: Any) -> Any:
        """
        Log *event* with **kw** at **critical** level.
        """

    def msg(self, event: str, **kw: Any) -> Any:
        """
        Log *event* with **kw** at **info** level.
        """



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import print_function
import platform
import sys
import threading
import contrast
from contrast import __version__, AGENT_CURR_WORKING_DIR
from contrast.reporting import teamserver_messages
from contrast.agent import (
    patch_controller,
    service_client,
    scope,
    thread_watcher,
)
from contrast.agent import agent_lib
from contrast.agent.protect import input_analysis
from contrast.agent.request_context import RequestContext
from contrast.agent.assess.rules.response.xss import analyze_xss
from contrast.agent.assess.rules.response.analyze import analyze_response_rules
from contrast.agent.assess.preflight import update_preflight_hashes
from contrast.agent.assess.rules.providers.enable import enable_providers
from contrast.agent.middlewares.route_coverage.routes_mixin import RoutesMixin
from contrast.agent.settings import Settings
from contrast.agent.speedracer_input_analysis import get_input_analysis
from contrast.api.attack import ProtectResponse
from contrast.reporting import ReportingClient
from contrast.reporting.activity_masker import ActivityMasker
from contrast.utils.decorators import cached_property
from contrast.utils.exceptions.contrast_service_exception import (
    ContrastServiceException,
)
from contrast.utils.decorators import log_time_cm
from contrast.utils.exceptions.security_exception import SecurityException
from contrast.utils.library_reader.library_reader import LibraryReader
from contrast.utils.loggers.logger import setup_agent_logger, setup_basic_agent_logger
from contrast.utils.timer import now_ms_float, now_ms
from contrast.utils.decorators import fail_loudly, fail_quietly
from contrast.assess_extensions import cs_str
from contrast.agent.protect.rule.sqli_rule import SqlInjection
from contrast.agent.protect.rule.cmdi_rule import CmdInjection
from contrast.agent.protect.rule.base_rule import BaseRule

if not contrast.telemetry_disabled():
    from contrast.agent.telemetry import Telemetry
else:
    Telemetry = None

# initialize a basic logger until config is parsed
logger = setup_basic_agent_logger()


class BaseMiddleware(RoutesMixin):
    """
    BaseMiddleware contains all the initial setup for the framework middlewares

    Requirements:

        1. It's callable
        2. It has call_with_agent
        3. It has call_without_agent

    Pre and post filter calls should not block the flow that this class has.

    Pre -> get_response -> post
    """

    _loaded = False
    app_name = ""  # This should be overridden by child classes

    DIAGNOSTIC_ENDPOINT = "/save-contrast-security-config"
    DIAGNOSTIC_ALLOWED_SERVER = "localhost"
    DIAGNOSTIC_ALLOWED_IP = "127.0.0.1"
    OVERRIDE_MESSAGE = (
        "A security filter prevented original response from being returned."
    )
    LOGS_SEPARATOR = "-" * 120

    def __init__(self):
        """
        If this method is called more than once per process, we use the _loaded
        flag to only run the following work once:
            - turning on service
            - library analysis thread initialization
            - turning on patches
            - hardcoded rule providers
            - scanning config rules

        the following work will be done every time:
            - any logging
            - attribute definition
            - loading common configuration
            - initializing settings
        """
        # id will be different across processes but also for multiple middlewares
        # within the same process
        self.id = id(self)
        self.settings = None
        self.first_request = True
        self.request_start_time = None
        self.request_path = None
        self.routes = {}

        if BaseMiddleware._loaded:
            logger.warning(
                "Contrast Agent middleware initialized more than once per process."
            )

        self.log_initialize()

        if not self.initialize_settings():
            return

        if not self.settings.is_agent_config_enabled():
            logger.warning("Contrast Agent is not enabled.")
            return

        setup_agent_logger(self.settings.config)
        self.settings.config.log_config()

        if not BaseMiddleware._loaded:
            cs_str.init_contrast_scope_cvars()

        if not BaseMiddleware._loaded and not self.initialize_service_client():
            logger.error("Unable to initialize Contrast Agent.")
            return

        # This call must happen after initializing SR so we have all data to log
        self.log_environment()
        self.warn_for_misleading_config()

        self.reporting_client = ReportingClient()
        self.reporting_client.start()

        if self.settings.config.is_service_bypassed:
            if not self.successful_startup_msgs():
                logger.error(
                    "Unable to initialize Contrast Agent.", direct_to_teamserver=1
                )
                return

        # This MUST happen after the initialization calls for Service and/or TeamServer
        # messaging (sending server start and application start) to ensure that
        # TeamServer will accept the messages sent by our background reporting threads
        thread_watcher.ensure_running(self)

        if not BaseMiddleware._loaded:
            self.initialize_libraries()

            patch_controller.enable_patches()

        if self.settings.is_assess_enabled():
            # For now agent runtime starts before config scanning
            # this will be reset when time_limit_threshold is reached,
            # it doesn't symbolize the total agent runtime for all time.
            self.settings.agent_runtime_window = now_ms()

            if not BaseMiddleware._loaded:
                enable_providers()
                self.scan_configs_in_thread()

        if (
            self.settings.config.is_agent_lib_enabled
            and self.settings.is_protect_enabled()
        ):
            agent_lib.initialize()
        else:
            logger.debug("Not using agent-lib")

        self.log_finish_initialize()
        # Starting telemetry must happen at the very end of initialization
        # so we have all the data to send in startup messages.
        if not BaseMiddleware._loaded and Telemetry is not None:
            contrast.TELEMETRY = Telemetry()
            contrast.TELEMETRY.start()

        BaseMiddleware._loaded = True

    @cached_property
    def name(self):
        raise NotImplementedError("Must implement name")

    def successful_startup_msgs(self):
        server_start_msg = teamserver_messages.ServerStart()
        response = self.reporting_client.send_message(server_start_msg)
        if response is None:
            return False

        server_start_msg.process_response(response, self.reporting_client)

        app_start_msg = teamserver_messages.ApplicationStartup()
        response = self.reporting_client.send_message(app_start_msg)
        if response is None:
            return False

        app_start_msg.process_response(response, self.reporting_client)
        return True

    @fail_loudly("Unable to initialize Contrast Agent Settings.", return_value=False)
    def initialize_settings(self):
        """
        Initialize agent settings.

        Returns True on settings being initialized and False if any failure.
        """
        self.settings = Settings(app_name=self.app_name, framework_name=self.name)
        return True

    def initialize_service_client(self):
        """
        Initialize connection to service.

        Returns True on service being initialized and False if any failure.
        """
        try:
            service_client.send_startup_messages()
        except Exception as e:
            # catches any exception including ContrastServiceException
            logger.error(e, exc_info=e)
            return False
        return True

    def warn_for_misleading_config(self):
        protect_enabled = self.settings.is_protect_enabled()
        assess_enabled = self.settings.is_assess_enabled()

        logger.info("Protect: %s", protect_enabled)
        logger.info("Assess: %s", assess_enabled)

        if protect_enabled and self.settings.config.get("assess.enable"):
            logger.warning("Protect is running but Assess is enabled in local config")
            logger.warning("Defaulting to Protect behavior only")

        if not protect_enabled and self.settings.config.get("protect.enable", True):
            logger.warning("Protect enabled in local config but disabled by Teamserver")

        if (
            self.settings.is_agent_config_enabled()
            and not protect_enabled
            and not assess_enabled
        ):
            logger.warning("Neither Protect nor Assess is running")

    def log_environment(self):
        """
        Log current working directory, python version and pip version
        """
        banner = f"{'-' * 50}ENVIRONMENT{'-' * 50}"
        logger.debug(banner)
        logger.debug("Current Working Dir: %s", AGENT_CURR_WORKING_DIR)
        logger.debug("Python Version: %s", sys.version)
        logger.debug("Framework Version: %s", self.settings.framework)
        logger.debug("Server Version: %s", self.settings.server)
        logger.debug("Contrast Python Agent Version: %s", __version__)
        logger.debug(
            "Contrast Service Version %s",
            self.settings.server_features.contrast_service,
        )
        logger.debug("Platform %s", platform.platform())

        try:
            import pip

            logger.debug("Pip Version: %s", pip.__version__)
        except Exception:
            pass

        logger.debug(banner)

    def initialize_libraries(self):
        """
        If enabled, read libraries from the application
        :return: True
        """
        if not self.settings.is_analyze_libs_enabled():
            return

        # Passing callback to send message due to a circular import issue, and a
        # deadlock occurring if something is imported inside of a function running in
        # this thread
        self.library_reader = LibraryReader(
            self.settings,
            send_sr_message_func=service_client.send_messages,
            send_ts_message_func=self.reporting_client.add_message,
        )
        self.library_reader.start_library_analysis_thread()

    def is_agent_enabled(self):
        """
        Agent is considered enabled if all of the following are true:
        1. config value for 'enable' is True (or empty, defaults to True)
            (meaning no errors during initialization process including initial
            connection to Speedracer)
        2. ServiceClient connection to Speedracer is True

        NOTE: If #3 is false (the connection to Speedracer is down at any time during
        the request cycle) then the agent is automatically disabled.
        """
        if self.settings is None:
            return False

        if not self.settings.is_agent_config_enabled():
            return False

        if not service_client.is_connected():
            try:
                service_client.send_startup_messages()
            except ContrastServiceException:
                return False

        return service_client.is_connected()

    def call_with_agent(self, *args):
        raise NotImplementedError("Must implement call_with_agent")

    def call_without_agent(self, *args):
        """
        The agent does not set context when this function is called so all other patches
        (e.g propagators) that check context shouldn't run.
        """
        logger.debug("Contrast Agent did not analyze request %s", self.request_path)

    def should_analyze_request(self, environ):
        """
        Determine if request should be analyzed based on configured settings.

        While returning different types of objects based on logic is not a good
        pattern, in this case it's an optimization that allows us to create
        the request context obj when we need it.

        :return: False or RequestContext instance
        """
        if not self.is_agent_enabled():
            logger.debug("Will not analyze request: agent disabled.")
            return False

        context = RequestContext(environ)

        path = environ.get("PATH_INFO")

        if self.settings.evaluate_exclusions(context, path):
            logger.debug("Will not analyze request: request meets exclusions.")
            return False

        from contrast.agent.assess import sampling

        if sampling.enabled() and sampling.meets_criteria(context):
            logger.debug("Will not analyze request: request meets sampling.")
            return False

        return context  # equivalent to returning True

    def handle_ensure(self, context, request):
        """
        Method that should run for all middlewares AFTER every request is made.
        """
        if context is None:
            logger.error("Context not defined in middleware ensure")
            return

        with scope.contrast_scope():
            thread_watcher.ensure_running(self)

            if self.settings.is_assess_enabled():
                # route discovery and storage needs to occur before sending messages
                self.handle_routes(context, request)

                logger.debug("Updating preflight hashes with route info")
                update_preflight_hashes(context)

            logger.debug("Sending final messages for reporting.")

            context.truncate_request_body()

            for msg in self.final_ts_messages(context):
                self.reporting_client.add_message(msg)

            # Mask sensitive data with activity masker
            # takes context.activity goes through http_request for activity
            # and masks body, cookies, headers, query and request parameters
            ActivityMasker(context).mask_sensitive_data()
            self.reporting_client.add_message(teamserver_messages.Activity(context))

            if not self.settings.config.is_service_bypassed:
                # messages still sent to SR until transition.
                final_messages = self.final_messages(context)
                service_client.send_messages(final_messages)

        self.first_request = False

    def final_ts_messages(self, context):
        """
        Messages sent to TS every time, regardless of bypass
        """
        messages = [
            teamserver_messages.Preflight(
                context.findings, context.activity.http_request
            ),
        ]

        # Currently we do not report an observed route if the route signature is empty.
        # As a team we've decided there isn't a meaningful default signature value
        # we can provide to customers. If a route doesn't show up in Contrast UI,
        # it may be due to its missing signature. In this scenario, we will have to work
        # with the customer directly to understand why the signature was not created.
        if self.settings.is_assess_enabled() and context.observed_route.signature:
            messages.append(teamserver_messages.ObservedRoute(context.observed_route))

        if self.first_request and self.settings.is_inventory_enabled():
            # ApplicationUpdate message only sent at first request once we have routes
            messages.append(teamserver_messages.ApplicationInventory(self.routes))

        return messages

    def final_messages(self, context):
        return [context.activity]

    def generate_security_exception_response(self):
        """
        Generate a response to be returned in the case of a SecurityException.

        Middlewares must override this method in order to provide the kind
        of response object that is expected by that framework.
        """
        raise NotImplementedError(
            "Middlewares must provide their own security response generator"
        )

    def handle_exception(self, exception):
        """
        Handle an exception being thrown especially if it is a SecurityException
        """
        logger.debug(
            "Handling %s raised in %s",
            exception.__class__.__name__,
            self.__class__.__name__,
        )
        if isinstance(exception, SecurityException):
            logger.info("%s: %s", exception.__class__.__name__, exception)
            logger.debug("Overriding response in Contrast Middleware")
            return self.generate_security_exception_response()

        logger.error("Reraising %r", exception)
        raise

    def prefilter(self, context):
        """
        Prefilter - AKA input analysis - is performed mostly in Speedracer but partly
        in the agent.

        This is PROTECT only.

        In this method we call on speedracer to do input analysis, which can result in:
        1. Speedracer finds an attack in which case we block the request
        2. Speedracer returns input analysis to use for later sink / infilter analysis,
            in which case we store it here in the request context.
        """
        if not self.settings.is_protect_enabled():
            return

        if (
            self.settings.config.is_agent_lib_enabled
            and self.settings.is_protect_enabled()
        ):
            input_analysis.get_input_analysis()

        with log_time_cm("protect prefilter"):
            context.speedracer_input_analysis = get_input_analysis()
            self.agent_prefilter()

    def agent_prefilter(self):
        """
        Prefilter for any rules that do not yet use speedracer
        """
        with scope.contrast_scope():
            self.prefilter_defend()

    @fail_quietly("Failed to run prefilter protect.")
    def prefilter_defend(self):
        rules = self.settings.protect_rules
        logger.debug("PROTECT: Running Agent prefilter.")

        for rule in rules.values():
            is_prefilter = rule.is_prefilter()

            if is_prefilter:
                rule.prefilter()

    def postfilter(self, context):
        """
        For all postfilter enabled rules.
        """
        with log_time_cm("postfilter"):
            if self.settings.is_protect_enabled():
                self.postfilter_defend(context)

            if self.settings.is_assess_enabled():
                with scope.contrast_scope():
                    self.response_analysis(context)

    def _process_trigger_handler(self, handler):
        """
        Gather metadata about response handler callback for xss trigger node

        We need to check whether the response handler callback is an instance method or
        not. This affects the way that our policy machinery works, and it also affects
        reporting, so we need to make sure to account for the possibility that handler
        is a method of some class rather than a standalone function.

        This should be called by the `trigger_node` method in child classes.
        """
        module = handler.__module__
        class_name = ""

        if hasattr(handler, "__self__"):
            class_name = handler.__self__.__class__.__name__
            args = (handler.__self__,)
            instance_method = True
        else:
            args = ()
            instance_method = False

        return module, class_name, args, instance_method

    @cached_property
    def trigger_node(self):
        """
        Trigger node property used by assess reflected xss postfilter rule

        This must be overridden by child classes that make use of the reflected
        xss postfilter rule.
        """
        raise NotImplementedError("Children must define trigger_node property")

    @fail_loudly("Unable to do assess response analysis")
    def response_analysis(self, context):
        """
        Run postfilter for any assess rules. Reflected xss rule runs by default.
        May be overridden in child classes.

        If the response content type matches a allowed content type, do not run
        assess xss response analysis. This is because the security team
        considers reflected xss within these content types to be a false positive.
        """
        logger.debug("ASSESS: Running response analysis")

        analyze_xss(context, self.trigger_node)

        if not self.settings.response_scanning_enabled:
            return

        if self.settings.config.is_service_bypassed:
            analyze_response_rules(context)

    @fail_quietly("Failed to run postfilter protect.")
    def postfilter_defend(self, context):
        rules = self.settings.protect_rules
        logger.debug("PROTECT: Running Agent postfilter.")
        if self.settings.config.is_agent_lib_enabled:
            context.user_input_analysis = []
            # Here we might have empty values if the rule is not enabled in settings
            rules_list = [
                (
                    "sql-injection",
                    self.settings.protect_rules.get(SqlInjection.RULE_NAME),
                ),
                (
                    "cmd-injection",
                    self.settings.protect_rules.get(CmdInjection.RULE_NAME),
                ),
            ]

            input_analysis.get_input_analysis(rules_list)
            if context.user_input_analysis:
                for rule in context.user_input_analysis[0]:
                    if rule.score >= 90:
                        BaseRule().build_or_append_attack(rule)
        for rule in rules.values():
            if rule.is_postfilter():
                rule.postfilter()

    def check_for_blocked(self, context):
        """
        Checks for BLOCK events in case SecurityException was caught by app code

        This should be called by each middleware after the view is generated
        but before returning the response (it can be before or after
        postfilter).

        If we make it to this call, it implies that either no SecurityException
        occurred, or if one did occur, it was caught by the application. If we
        find a BLOCK here, it necessarily implies that an attack was detected
        in the application, but the application caught our exception. If the
        application hadn't caught our exception, we never would have made it
        this far because the exception would have already bubbled up to the
        middleware exception handler. So this is really our first and our last
        opportunity to check for this particular edge case.
        """
        for attack in context.attacks:
            if attack.response == ProtectResponse.BLOCKED:
                msg = f"Rule {attack.rule_id} threw a security exception"
                raise SecurityException(None, message=msg)

    def log_start_request_analysis(self):
        self.request_start_time = now_ms_float()

        logger.debug("Beginning request analysis", request_path=self.request_path)

    def log_end_request_analysis(self):
        request_end_time = now_ms_float()

        logger.debug(
            "Ending request analysis",
            request_path=self.request_path,
        )

        elapsed_time = request_end_time - self.request_start_time

        logger.debug(
            "elapsed time request analysis ms",
            elapsed_time=elapsed_time,
            request_path=self.request_path,
        )

    def log_initialize(self):
        logger.info(
            "Initializing Contrast Agent %s [id=%s]", self.__class__.__name__, self.id
        )

        logger.info("Contrast Python Agent Version: %s\n", __version__)

    def log_finish_initialize(self):
        logger.info(
            "Finished Initializing Contrast Agent %s [id=%s] \n\n%s\n",
            self.__class__.__name__,
            self.id,
            self.LOGS_SEPARATOR,
        )

    def scan_configs_in_thread(self):
        """
        Run config scanning rules for assess in a separate thread.

        Not all frameworks we support will necessarily have config scanning rules.
        These frameworks should not start a thread.
        """
        if hasattr(self, "_scan_configs"):
            logger.debug("Will start a thread to scan config rules.")
            scanner = threading.Thread(target=self._scan_configs)
            scanner.daemon = True
            scanner.start()
        else:
            logger.debug(f"No config scanning rules for {self.__class__.__name__}")



# SPDX-License-Identifier: MIT OR Apache-2.0
# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the MIT License.  See the LICENSE file in the root of this
# repository for complete details.

"""
Logger wrapper and helper class.
"""

from typing import Any, Dict, Iterable, Mapping, Optional, Sequence, Tuple

from contrast.extern.structlog.exceptions import DropEvent

from .types import BindableLogger, Context, Processor, WrappedLogger


class BoundLoggerBase:
    """
    Immutable context carrier.

    Doesn't do any actual logging; examples for useful subclasses are:

    - the generic `BoundLogger` that can wrap anything,
    - `structlog.stdlib.BoundLogger`.
    - `structlog.twisted.BoundLogger`,

    See also `custom-wrappers`.
    """

    _logger: WrappedLogger
    """
    Wrapped logger.

    .. note::

        Despite underscore available **read-only** to custom wrapper classes.

        See also `custom-wrappers`.
    """

    def __init__(
        self,
        logger: WrappedLogger,
        processors: Iterable[Processor],
        context: Context,
    ):
        self._logger = logger
        self._processors = processors
        self._context = context

    def __repr__(self) -> str:
        return "<{}(context={!r}, processors={!r})>".format(
            self.__class__.__name__, self._context, self._processors
        )

    def __eq__(self, other: Any) -> bool:
        try:
            if self._context == other._context:
                return True
            else:
                return False
        except AttributeError:
            return False

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def bind(self, **new_values: Any) -> "BoundLoggerBase":
        """
        Return a new logger with *new_values* added to the existing ones.
        """
        return self.__class__(
            self._logger,
            self._processors,
            self._context.__class__(self._context, **new_values),
        )

    def unbind(self, *keys: str) -> "BoundLoggerBase":
        """
        Return a new logger with *keys* removed from the context.

        :raises KeyError: If the key is not part of the context.
        """
        bl = self.bind()
        for key in keys:
            del bl._context[key]

        return bl

    def try_unbind(self, *keys: str) -> "BoundLoggerBase":
        """
        Like :meth:`unbind`, but best effort: missing keys are ignored.

        .. versionadded:: 18.2.0
        """
        bl = self.bind()
        for key in keys:
            bl._context.pop(key, None)

        return bl

    def new(self, **new_values: Any) -> "BoundLoggerBase":
        """
        Clear context and binds *initial_values* using `bind`.

        Only necessary with dict implementations that keep global state like
        those wrapped by `structlog.threadlocal.wrap_dict` when threads
        are re-used.
        """
        self._context.clear()

        return self.bind(**new_values)

    # Helper methods for sub-classing concrete BoundLoggers.

    def _process_event(
        self, method_name: str, event: Optional[str], event_kw: Dict[str, Any]
    ) -> Tuple[Sequence[Any], Mapping[str, Any]]:
        """
        Combines creates an ``event_dict`` and runs the chain.

        Call it to combine your *event* and *context* into an event_dict and
        process using the processor chain.

        :param method_name: The name of the logger method.  Is passed into
            the processors.
        :param event: The event -- usually the first positional argument to a
            logger.
        :param event_kw: Additional event keywords.  For example if someone
            calls ``log.msg("foo", bar=42)``, *event* would to be ``"foo"``
            and *event_kw* ``{"bar": 42}``.

        :raises: `structlog.DropEvent` if log entry should be dropped.
        :raises: `ValueError` if the final processor doesn't return a
            str, bytes, bytearray, tuple, or a dict.

        :returns: `tuple` of ``(*args, **kw)``

        .. note::

            Despite underscore available to custom wrapper classes.

            See also `custom-wrappers`.

        .. versionchanged:: 14.0.0
            Allow final processor to return a `dict`.
        .. versionchanged:: 20.2.0
            Allow final processor to return `bytes`.
        .. versionchanged:: 21.2.0
            Allow final processor to return a `bytearray`.
        """
        # We're typing it as Any, because processors can return more than an
        # EventDict.
        event_dict: Any = self._context.copy()
        event_dict.update(**event_kw)

        if event is not None:
            event_dict["event"] = event
        for proc in self._processors:
            event_dict = proc(self._logger, method_name, event_dict)

        if isinstance(event_dict, (str, bytes, bytearray)):
            return (event_dict,), {}
        elif isinstance(event_dict, tuple):
            # In this case we assume that the last processor returned a tuple
            # of ``(args, kwargs)`` and pass it right through.
            return event_dict  # type: ignore
        elif isinstance(event_dict, dict):
            return (), event_dict
        else:
            raise ValueError(
                "Last processor didn't return an appropriate value.  Valid "
                "return values are a dict, a tuple of (args, kwargs), bytes, "
                "or a str."
            )

    def _proxy_to_logger(
        self, method_name: str, event: Optional[str] = None, **event_kw: Any
    ) -> Any:
        """
        Run processor chain on event & call *method_name* on wrapped logger.

        DRY convenience method that runs :func:`_process_event`, takes care of
        handling :exc:`structlog.DropEvent`, and finally calls *method_name* on
        :attr:`_logger` with the result.

        :param method_name: The name of the method that's going to get
            called.  Technically it should be identical to the method the
            user called because it also get passed into processors.
        :param event: The event -- usually the first positional argument to a
            logger.
        :param event_kw: Additional event keywords.  For example if someone
            calls ``log.msg("foo", bar=42)``, *event* would to be ``"foo"``
            and *event_kw* ``{"bar": 42}``.

        .. note::

            Despite underscore available to custom wrapper classes.

            See also `custom-wrappers`.
        """
        try:
            args, kw = self._process_event(method_name, event, event_kw)
            return getattr(self._logger, method_name)(*args, **kw)
        except DropEvent:
            return


def get_context(bound_logger: BindableLogger) -> Context:
    """
    Return *bound_logger*'s context.

    The type of *bound_logger* and the type returned depend on your
    configuration.

    :param bound_logger: The bound logger whose context you want.

    :returns: The *actual* context from *bound_logger*. It is *not* copied
        first.

    .. versionadded:: 20.2
    """
    # This probably will get more complicated in the future.
    return bound_logger._context



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import pathlib
import platform
import psutil
import time
from os import path

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


def make_sure_path_exists(log_path):
    directory = path.dirname(log_path)

    # log_path is not a dir, just a file path
    if directory in ("", "/"):
        return

    pathlib.Path(directory).mkdir(parents=True, exist_ok=True)


def sleep(seconds):
    """
    Wrapper around time.sleep that makes testing less complicated
    """
    time.sleep(seconds)


class ServiceUtil(object):
    SERVICE_EXE_NAME = "contrast-service"
    BASE_SERVICE_DIRECTORY = "service_executables"

    LINUX_PLATFORM = "linux"
    MAC_PLATFORM = "darwin"

    LINUX_SERVICE_DIRECTORY = LINUX_PLATFORM
    MAC_SERVICE_DIRECTORY = "mac"

    SUPPORTED_PLATFORM_DIRECTORIES = {
        LINUX_PLATFORM: LINUX_SERVICE_DIRECTORY,
        MAC_PLATFORM: MAC_SERVICE_DIRECTORY,
    }

    def __init__(self, service_log_path):
        self.service_log_path = service_log_path or "contrast_service.log"
        self.service_path = ServiceUtil.generate_service_path()

    def start_bundled_service(self):
        """
        Start a new Contrast Service. If one is already running on the host (bundled or external),
        for now we accept that there will be multiple services.

        :return: bool of success
        """
        return self._start_service()

    @property
    def service_exists(self):
        return self.service_path and path.exists(self.service_path)

    def _start_service(self):
        """
        Based on the operating system architecture, start the correct type of service.
        :return: bool of success
        """
        if self.service_exists:
            if self.service_log_path == "STDOUT":
                psutil.Popen(self.service_path)
            else:
                make_sure_path_exists(self.service_log_path)
                with open(self.service_log_path, "wb") as out:
                    psutil.Popen(self.service_path, stdout=out, stderr=out)

            logger.debug("Started up %s", self.SERVICE_EXE_NAME)
            return True

        logger.error("Failed to find service executable at %s", self.service_path)
        return False

    @staticmethod
    def generate_service_path():
        """
        Gets the absolute path to the appropriate service, depending on the current system architecture.
        """
        current_file_path = path.dirname(path.realpath(__file__))
        parent_directory = path.dirname(current_file_path)
        base_service_directory = path.realpath(
            path.join(parent_directory, ServiceUtil.BASE_SERVICE_DIRECTORY)
        )
        platform_directory = ServiceUtil.get_platform_directory()
        exe_name = ServiceUtil.SERVICE_EXE_NAME
        return path.join(base_service_directory, platform_directory, exe_name)

    @staticmethod
    def get_platform_directory():
        """
        Gets the platform directory name corresponding to the current system architecture.
        For an unknown system, default to Linux.
        :return: A string of the directory name
        """
        platform_directory = ServiceUtil.SUPPORTED_PLATFORM_DIRECTORIES.get(
            platform.system().lower()
        )
        if platform_directory is None:
            logger.debug(
                "Unknown/unsupported host platform; will attempt to use %s %s"
                " executable",
                ServiceUtil.LINUX_PLATFORM,
                ServiceUtil.SERVICE_EXE_NAME,
            )
            return ServiceUtil.LINUX_SERVICE_DIRECTORY
        return platform_directory



# SPDX-License-Identifier: MIT OR Apache-2.0
# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the MIT License.  See the LICENSE file in the root of this
# repository for complete details.

"""
Structured Logging for Python
"""

from contrast.extern.structlog import dev, processors, stdlib, testing, threadlocal, types
from contrast.extern.structlog._base import BoundLoggerBase, get_context
from contrast.extern.structlog._config import (
    configure,
    configure_once,
    get_config,
    get_logger,
    getLogger,
    is_configured,
    reset_defaults,
    wrap_logger,
)
from contrast.extern.structlog._generic import BoundLogger
from contrast.extern.structlog._log_levels import make_filtering_bound_logger
from contrast.extern.structlog._loggers import (
    BytesLogger,
    BytesLoggerFactory,
    PrintLogger,
    PrintLoggerFactory,
)
from contrast.extern.structlog.exceptions import DropEvent
from contrast.extern.structlog.testing import ReturnLogger, ReturnLoggerFactory


try:
    from contrast.extern.structlog import twisted
except ImportError:
    twisted = None  # type: ignore

try:
    from contrast.extern.structlog import contextvars
except ImportError:
    contextvars = None  # type: ignore


__version__ = "21.5.0"

__title__ = "structlog"
# __doc__ is None when running with PYTHONOPTIMIZE=2 / -OO
__description__ = (__doc__ or "").strip()

__uri__ = "https://www.structlog.org/"

__author__ = "Hynek Schlawack"
__email__ = "hs@ox.cx"

__license__ = "MIT or Apache License, Version 2.0"
__copyright__ = "Copyright (c) 2013 " + __author__


__all__ = [
    "BoundLogger",
    "BoundLoggerBase",
    "BytesLogger",
    "BytesLoggerFactory",
    "DropEvent",
    "PrintLogger",
    "PrintLoggerFactory",
    "ReturnLogger",
    "ReturnLoggerFactory",
    "configure",
    "configure_once",
    "contextvars",
    "dev",
    "getLogger",
    "get_config",
    "get_context",
    "get_logger",
    "is_configured",
    "make_filtering_bound_logger",
    "processors",
    "reset_defaults",
    "stdlib",
    "testing",
    "threadlocal",
    "twisted",
    "types",
    "wrap_logger",
]



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
class UserInput:
    PARAMETER_VALUE = 6
    QUERYSTRING = 7
    METHOD = 22
    UNKNOWN = 99

    _TYPE_TO_STR = {
        PARAMETER_VALUE: "PARAMETER_VALUE",
        QUERYSTRING: "QUERYSTRING",
        METHOD: "METHOD",
        UNKNOWN: "UNKNOWN",
    }

    def __init__(self, input_type, key, value, path="", matcher_ids=None,
                 document_type=0):
        self.input_type = input_type
        self.key = key
        self.value = value
        self.path = path
        self.matcher_ids = [] if matcher_ids is None else matcher_ids

        self.document_type = document_type

    @classmethod
    def type_to_str(cls, intput_type: int) -> str:
        """
        Convert int to str input type
        """
        return cls._TYPE_TO_STR.get(intput_type, "")




# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
class ContrastServiceException(Exception):
    def __init__(self, message=None):
        message = (
            "Unable to connect or send messages to Contrast Service."
            if not message
            else message
        )
        super().__init__(message)
